/**
 * Copyright (c) 2004 Craig Setera
 * All Rights Reserved.
 * Licensed under the Academic Free License version 1.2
 * For more information see http://www.opensource.org/licenses/academic.php
 */
package zmachine.ui.graphics;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;

import zmachine.util.StringTokenizer;

/**
 * The output console CustomItem implementation.
 * <p />
 * Copyright (c) 2003-2004 Craig Setera<br>
 * All Rights Reserved.<br>
 * Licensed under the Academic Free License version 1.2<p/>
 * <br>
 * $Revision: 1.1 $
 * <br>
 * $Date: 2004/08/21 19:32:11 $
 * <br>
 * @author Craig Setera
 */
public class ConsoleItem extends CustomItem {

	// Current colors in ZMachine numbers
	private int backColor;
	private char[] blankLine;
	private int charsPerLine;
	
	// The current font
	private Font currentFont;
	
	// Current display position
	private int[] cursorPosition;
	private int foreColor;
	private char[][] grid;
	private int height;
	
	// Working area for the display information
	private int lineCount;
	private String text;
	private int width;
	
	/**
	 * Constructor for a new ConsoleItem instance.
	 */
	ConsoleItem() {
		super(null);
		calculateBounds();
		
		currentFont = Font.getFont(
			Font.FACE_MONOSPACE, 
			Font.STYLE_PLAIN, 
			Font.SIZE_SMALL);
		
		// Calculate the metrics of the screen given
		// the font
		charsPerLine = (width / currentFont.charWidth('M')) - 1;
		lineCount = height / currentFont.getHeight();
		
		// Initialize the blank line
		blankLine = new char[charsPerLine];
		for (int i = 0; i < blankLine.length; i++) {
			blankLine[i] = ' ';
		}
		
		// Create the grid
		grid = new char[lineCount][];
		for (int i = 0; i < lineCount; i++) {
			grid[i] = new char[charsPerLine];
			eraseLine(i);
		}
		
	}
	
	/**
	 * Calculate the bounds of this item. 
	 */
	private void calculateBounds() {
		Canvas c = new Canvas() {
			protected void paint(Graphics arg0) {
			}
		};
		
		width = c.getWidth();
		height = c.getHeight();
		
		// Steal the space that the text entry
		// field should be from the height of 
		// this item
		Font f = Font.getDefaultFont();
		height -= (f.getHeight() * 2);
	}
	
	/**
	 * Copy the specified line contents into the target line.
	 * 
	 * @param srcLine
	 * @param tgtLine
	 */
	private void copyLine(int srcLine, int tgtLine) {
		copyLineChars(grid[srcLine], grid[tgtLine]);
	}
	
	/**
	 * Copy the line characters from one array to another,
	 * taking into account different array lengths.
	 * 
	 * @param src
	 * @param tgt
	 */
	private void copyLineChars(char[] src, char[] tgt) {
		int length = (src.length > tgt.length) ?
			tgt.length : src.length;
			
		System.arraycopy(src, 0, tgt, 0, length);
	}

	/**
	 * Move to the next line 
	 */
	void nextLine() {
		cursorPosition[0] = 0;
		
		// Scroll as necessary
		if ((cursorPosition[1] + 1) >= lineCount) {
			scrollWindow(1);
		} else {
			cursorPosition[1]++;
		}

		// Decide if we need to pause
//		if ((linesOutputSinceInput + 1) >= lineCount) {
//			pause();
//		} else {
//			linesOutputSinceInput++;
//		}
	}

	/**
	 * Output the specified text without further breaking it down.
	 * 
	 * @param text
	 * @return
	 */
	private void outputTextNoBreak(String text)
	{
		if (text.length() > 0) {
			if ((cursorPosition[0] != 0) || (text.charAt(0) != ' ')) { 
				int textLength = text.length();
				
				for (int i = 0; i < textLength; i++) {
					setCharacterAt(
						cursorPosition[1], 
						cursorPosition[0] + i, 
						text.charAt(i));
				}
				
				cursorPosition[0] += textLength;
			}
		}
	}

	/**
	 * Pause the output and wait before scrolling at all... 
	 */
	private void pause() {
		// TODO Auto-generated method stub
		
	}

	/**
	 * Scroll the current grid up the specified number of lines.
	 * 
	 * @param numLines
	 */
	private void scrollUp(int numLines) {
		for (int i = numLines; i < lineCount; i++) {
			copyLine(i, i - numLines);
		}
		eraseLine(lineCount - 1);
	}

	/**
	 * Set the specified grid element.
	 * 
	 * @param row
	 * @param col
	 * @param c
	 */
	private void setCharacterAt(int row, int col, char c) {
		try {
			grid[row][col] = c;
		} catch (ArrayIndexOutOfBoundsException e) {
			System.out.println("Index (" + row + ", " + col + ") out of bounds for '" + c + "'");
		}
	}
	
	/**
	 * Output the specified line of text to the screen, possibly
	 * doing word wrapping to fit the screen.
	 * 
	 * @param line
	 */
	void showLine(String text) {
		// See if we can fit text on to the current line
		if ((cursorPosition[0] + text.length()) < charsPerLine) {
			outputTextNoBreak(text);
		} else {
			// Nope... break up the string and operate on wordwrap
			StringTokenizer st = new StringTokenizer(text, "\t ", true);
			while (st.hasMoreTokens()) {
				String word = st.nextToken();
				if ((cursorPosition[0] + word.length()) >= charsPerLine) {
					nextLine();
				}
				
				outputTextNoBreak(word);
			}
		}
	}
	
	/**
	 * @see zmachine.ui.CustomItem#getMinContentHeight()
	 */
	protected int getMinContentHeight() {
		return height;
	}

	/**
	 * @see zmachine.ui.CustomItem#getMinContentWidth()
	 */
	protected int getMinContentWidth() {
		return width;
	}

	/**
	 * @see zmachine.ui.CustomItem#getPrefContentHeight()
	 */
	protected int getPrefContentHeight() {
		return height;
	}

	/**
	 * @see zmachine.ui.CustomItem#getPrefContentWidth()
	 */
	protected int getPrefContentWidth() {
		return width;
	}

	/**
	 * @see zmachine.ui.CustomItem#paint(javax.microedition.lcdui.Graphics, int, int)
	 */
	protected void paint(Graphics graphics, int w, int h) {
		int charHeight = currentFont.getHeight();
		graphics.setFont(currentFont);
		
		graphics.setColor(255,255,255);
		graphics.fillRect(0, 0, w, h);
		graphics.setColor(0, 0, 0);
		
		for (int i = 0; i < lineCount; i++) {
			graphics.drawChars(
				grid[i], 
				0,
				grid[i].length,
				0, 
				(i * charHeight), 
				Graphics.TOP | Graphics.LEFT);
		}
	}

	/**
	 * @see zmachine.ZUserInterface#eraseLine(int)
	 */
	void eraseLine(int line) {
		copyLineChars(blankLine, grid[line]);
	}

	/**
	 * @see zmachine.ZUserInterface#eraseWindow(int)
	 */
	void eraseWindow(int window) {
		for (int i = 0; i < lineCount; i++) {
			eraseLine(i);
		}
	}

	/**
	 * @see zmachine.ZUserInterface#getCursorPosition()
	 */
	int[] getCursorPosition() {
		return cursorPosition;
	}

	/**
	 * @see zmachine.ZUserInterface#getFontSize()
	 */
	int[] getFontSize() {
		int width = currentFont.charWidth(' ');
		int height = currentFont.getHeight();
		
		return new int[] { width, height };
	}

	/**
	 * @see zmachine.ZUserInterface#getScreenCharacters()
	 */
	int[] getScreenCharacters() {
		return new int[] { charsPerLine, lineCount };
	}

	/**
	 * @see zmachine.ZUserInterface#getScreenUnits()
	 */
	int[] getScreenUnits() {
		int[] units = getFontSize();
		units[0] *= charsPerLine;
		units[1] *= lineCount;
		
		return units;
	}

	/**
	 * @see zmachine.ZUserInterface#getWindowSize(int)
	 */
	int[] getWindowSize(int window) {
		// TODO This is not really correct
		return getScreenCharacters();
	}

	/**
	 * @see zmachine.ZUserInterface#initialize(int)
	 */
	public void initialize(int ver) {
		cursorPosition = new int[] { 0, 0 };
	}

	/**
	 * @see zmachine.ZUserInterface#scrollWindow(int)
	 */
	void scrollWindow(int lines) {
		if (lines < 0) {
			// TODO Fix this...
			System.out.println("scrollWindow down not implemented");
		} else {
			scrollUp(lines);
		}
	}

	/**
	 * @see zmachine.ZUserInterface#setColor(int, int)
	 */
	void setColor(int fg, int bg) {
		foreColor = fg;
		backColor = bg;
	}

	/**
	 * @see zmachine.ZUserInterface#setCursorPosition(int, int)
	 */
	void setCursorPosition(int x, int y) {
		cursorPosition[0] = x;
		cursorPosition[1] = y;
	}

	/**
	 * @see zmachine.ZUserInterface#showString(java.lang.String)
	 */
	void showString(String s) {
		try {
			// First, break up the string into lines based on
			// terminating characters
			StringTokenizer lines = new StringTokenizer(s, "\n", false);
			while (lines.hasMoreTokens()) {
				String line = lines.nextToken();
				showLine(line);
				
				if (lines.hasMoreTokens()) {
					nextLine();
				}
			}
		} catch (Throwable t) {
			t.printStackTrace();
		}
		
		// Time to repaint
		repaint();
	}

}
